// There are 73 different Nmea0183 sentences. Each sentence is defined by 3 ASCII characters
// The 1st ASCII character, A, is hexadecimal 41 or binary 01000001
// The 2nd ASCII character, B, is hexadecimal 42 or binary 01000010
// The 3rd ASCII character, C, is hexadecimal 43 or binary 01000011
// The 4th ASCII character, D, is hexadecimal 44 or binary 01000100
// ....
// The end ASCII character, Z, is hexadecimal 5A or binary 01011010
// So only the least 5 significative bits suffice to define each character
// as the 3 most significative bits are always 010. Therefore 15 bits
// are enough to difine the string of 3 letters that define a Nmea sentence. The
// sentence ZDA, for example will be 11010 00100 00001. If a sentence ZZZ existed
// it could be represented by 11010 11010 11010 which in decimal is 27482

// So we will have a table called NmeaTable[] with 27483 entries. Each entry
// corresponds to a string of 3 capital letters, that is a unique Nmea sentence.
// But if we want to filter 7 channels we will not want to have 7 tables with
// 27483 entries to hold 0 if the sentence is to reject or 1 if it is to accept.
// The majority of the entries will not corresponf to a valid sentence.
// Ideally we would like to have only 7 tables with 74 entries! So we will have
// 7 tables with 256 entries each having space for possible expansion.
// Now look to this array:

const char *NMEA183[] = { "AAM", "ALM", "APA", "BOD", "BWC", "BWR", "BWW", "DBK", "DBS", "DBT", "DCN", 
                          "DPT", "DTM", "FSI", "GBS", "GGA", "GLC", "GLL", "GNS", "GRS", "GST", "GSA",
                          "GSV", "GTD", "GXA", "HDG", "HDM", "HDT", "HFB", "HSC", "ITS", "LCD", "MSK", 
                          "MSS", "MTW", "MWV", "OLN", "OSD", "ROO", "RMA", "RMB", "RMC", "ROT", "RPM", "RSA", 
                          "RSD", "RTE", "SFI", "STN", "TDS", "TFI", "TPC", "TPR", "TRF", "TTM", "VBW", 
                          "VDR", "VHW", "VLW", "VPW", "VTG", "VWR", "WCV", "WNC", "WPL", "XDR", "XTE", 
                          "XTR", "ZDA", "ZFO", "ZTG", "VDM", "VDO"};

// AAM will be the 1th sentence, ALM will be the 2th sentence, and ... VDO will be the 74th sentence
// But AAM is 00001 00001 01101 or in decimal 1069. So the big NmeaTable[27483] will be all filled with 
// zeros and in the 1069 th position will have 1. The next sentence ALM is 00001 01100 01101 or 1421 in
// in decimal. So in the big table with all zeros we put 2 in position 1421. And we will write 74 in
// the position corresponding to VDO which. 

// Now for each of the 7 smaller tables we use also fill them with zeros and then using the FILTERING STRING
// change its content with 1 if the sentence is to be accepted. 

// In real time and given a sentence, we build the 15 bit number formed by shifting 10 positions the 5 bit that
// corresponds to the 1st character, shifting the 2nd character by 5 bits and adding these 2 shifted values to
// the 3rd character. Using that resulting number we look to the BIG table to see which one of the 74 sentences we
// are dealing with. Then we just use that number to get a 0 or a 1 in the Table for the channel we want to filter. 


int NMEA183Size;

// these parameters are for nmea sentence filtering

byte NmeaTable[27483];
byte NmeaTab_1[256];    // for nmea0183 input P1
byte NmeaTab_3[256];    // for nmea0183 output P3
byte NmeaTab_4[256];    // for STi input
byte NmeaTab_5[256];    // for STo output

void BuildNmeaTables() {
  int i;
  String st;
  NMEA183Size = sizeof(NMEA183)/4;
  for ( i = 0; i < 27483; i++) {
    NmeaTable[i] = 0;
  }
  for ( i = 0; i < 256; i++) {
    NmeaTab_1[i] = 0;
    NmeaTab_3[i] = 0;
    NmeaTab_4[i] = 0;  
    NmeaTab_5[i] = 0;           
  }
  for ( i = 0; i < NMEA183Size; i++) {
    st = NMEA183[i];
    int a = st.charAt(0) & B00011111;
    int b = st.charAt(1) & B00011111;  
    int c = st.charAt(2) & B00011111; 
    a = a << 10;
    b = b << 5;
    a = a + b + c;
    NmeaTable[a] = i + 1;
  }
}

void SetNmeaTab_1(){
  int len, i; String nmea; 
  String st = String(FltStr_1); 
  len = st.length();  boolean Accept;
  FltSkip_1 = FltStr_1[0] - 49;
  if ( len == 1 ) { InitNmeaTab( 1, 1 ); return; }
  if ( FltStr_1[2] == 65 ) { InitNmeaTab( 1, 0 ); Accept = true; }
  if ( FltStr_1[2] == 82 ) { InitNmeaTab( 1, 1 ); Accept = false; }
  if ( len >  6 ) { nmea = st.substring( 4,7 ); i = GetNmeaIndByStr(nmea); NmeaTab_1[i] = Accept; } 
  if ( len > 10 ) { nmea = st.substring( 8,11); i = GetNmeaIndByStr(nmea); NmeaTab_1[i] = Accept; } 
  if ( len > 14 ) { nmea = st.substring(12,15); i = GetNmeaIndByStr(nmea); NmeaTab_1[i] = Accept; } 
  if ( len > 18 ) { nmea = st.substring(16,19); i = GetNmeaIndByStr(nmea); NmeaTab_1[i] = Accept; } 
  if ( len > 22 ) { nmea = st.substring(20,23); i = GetNmeaIndByStr(nmea); NmeaTab_1[i] = Accept; } 
  if ( len > 26 ) { nmea = st.substring(24,27); i = GetNmeaIndByStr(nmea); NmeaTab_1[i] = Accept; } 
  if ( len > 30 ) { nmea = st.substring(28,31); i = GetNmeaIndByStr(nmea); NmeaTab_1[i] = Accept; } 
  if ( len > 34 ) { nmea = st.substring(32,35); i = GetNmeaIndByStr(nmea); NmeaTab_1[i] = Accept; } 
  if ( len > 38 ) { nmea = st.substring(36,39); i = GetNmeaIndByStr(nmea); NmeaTab_1[i] = Accept; } 
  if ( len > 42 ) { nmea = st.substring(40,43); i = GetNmeaIndByStr(nmea); NmeaTab_1[i] = Accept; }  
}

void SetNmeaTab_3(){
  int len, i; String nmea; 
  String st = String(FltStr_3); 
  len = st.length();  boolean Accept;
  FltSkip_3 = FltStr_3[0] - 49;
  if ( len == 1 ) { InitNmeaTab( 3, 1 ); return; }
  if ( FltStr_3[2] == 65 ) { InitNmeaTab( 3, 0 ); Accept = true; }
  if ( FltStr_3[2] == 82 ) { InitNmeaTab( 3, 1 ); Accept = false; }
  if ( len >  6 ) { nmea = st.substring( 4,7 ); i = GetNmeaIndByStr(nmea); NmeaTab_3[i] = Accept; } 
  if ( len > 10 ) { nmea = st.substring( 8,11); i = GetNmeaIndByStr(nmea); NmeaTab_3[i] = Accept; } 
  if ( len > 14 ) { nmea = st.substring(12,15); i = GetNmeaIndByStr(nmea); NmeaTab_3[i] = Accept; } 
  if ( len > 18 ) { nmea = st.substring(16,19); i = GetNmeaIndByStr(nmea); NmeaTab_3[i] = Accept; } 
  if ( len > 22 ) { nmea = st.substring(20,23); i = GetNmeaIndByStr(nmea); NmeaTab_3[i] = Accept; } 
  if ( len > 26 ) { nmea = st.substring(24,27); i = GetNmeaIndByStr(nmea); NmeaTab_3[i] = Accept; } 
  if ( len > 30 ) { nmea = st.substring(28,31); i = GetNmeaIndByStr(nmea); NmeaTab_3[i] = Accept; } 
  if ( len > 34 ) { nmea = st.substring(32,35); i = GetNmeaIndByStr(nmea); NmeaTab_3[i] = Accept; } 
  if ( len > 38 ) { nmea = st.substring(36,39); i = GetNmeaIndByStr(nmea); NmeaTab_3[i] = Accept; } 
  if ( len > 42 ) { nmea = st.substring(40,43); i = GetNmeaIndByStr(nmea); NmeaTab_3[i] = Accept; }   
}

void SetNmeaTab_4(){
  int len, i; String nmea; 
  String st = String(FltStr_4); 
  len = st.length();  boolean Accept;
  FltSkip_4 = FltStr_4[0] - 49;
  if ( len == 1 ) { InitNmeaTab( 4, 1 ); return; }
  if ( FltStr_4[2] == 65 ) { InitNmeaTab( 4, 0 ); Accept = true; }
  if ( FltStr_4[2] == 82 ) { InitNmeaTab( 4, 1 ); Accept = false; }
  if ( len >  6 ) { nmea = st.substring( 4,7 ); i = GetNmeaIndByStr(nmea); NmeaTab_4[i] = Accept; } 
  if ( len > 10 ) { nmea = st.substring( 8,11); i = GetNmeaIndByStr(nmea); NmeaTab_4[i] = Accept; } 
  if ( len > 14 ) { nmea = st.substring(12,15); i = GetNmeaIndByStr(nmea); NmeaTab_4[i] = Accept; } 
  if ( len > 18 ) { nmea = st.substring(16,19); i = GetNmeaIndByStr(nmea); NmeaTab_4[i] = Accept; } 
  if ( len > 22 ) { nmea = st.substring(20,23); i = GetNmeaIndByStr(nmea); NmeaTab_4[i] = Accept; } 
  if ( len > 26 ) { nmea = st.substring(24,27); i = GetNmeaIndByStr(nmea); NmeaTab_4[i] = Accept; } 
  if ( len > 30 ) { nmea = st.substring(28,31); i = GetNmeaIndByStr(nmea); NmeaTab_4[i] = Accept; } 
  if ( len > 34 ) { nmea = st.substring(32,35); i = GetNmeaIndByStr(nmea); NmeaTab_4[i] = Accept; } 
  if ( len > 38 ) { nmea = st.substring(36,39); i = GetNmeaIndByStr(nmea); NmeaTab_4[i] = Accept; } 
  if ( len > 42 ) { nmea = st.substring(40,43); i = GetNmeaIndByStr(nmea); NmeaTab_4[i] = Accept; }    
}

void SetNmeaTab_5(){
  int len, i; String nmea; 
  String st = String(FltStr_5); 
  len = st.length();  boolean Accept;
  FltSkip_5 = FltStr_5[0] - 49;
  if ( len == 1 ) { InitNmeaTab( 5, 1 ); return; } 
  if ( FltStr_5[2] == 65 ) { InitNmeaTab( 5, 0 ); Accept = true; }
  if ( FltStr_5[2] == 82 ) { InitNmeaTab( 5, 1 ); Accept = false; }
  if ( len >  6 ) { nmea = st.substring( 4,7 ); i = GetNmeaIndByStr(nmea); NmeaTab_5[i] = Accept; } 
  if ( len > 10 ) { nmea = st.substring( 8,11); i = GetNmeaIndByStr(nmea); NmeaTab_5[i] = Accept; } 
  if ( len > 14 ) { nmea = st.substring(12,15); i = GetNmeaIndByStr(nmea); NmeaTab_5[i] = Accept; } 
  if ( len > 18 ) { nmea = st.substring(16,19); i = GetNmeaIndByStr(nmea); NmeaTab_5[i] = Accept; } 
  if ( len > 22 ) { nmea = st.substring(20,23); i = GetNmeaIndByStr(nmea); NmeaTab_5[i] = Accept; } 
  if ( len > 26 ) { nmea = st.substring(24,27); i = GetNmeaIndByStr(nmea); NmeaTab_5[i] = Accept; } 
  if ( len > 30 ) { nmea = st.substring(28,31); i = GetNmeaIndByStr(nmea); NmeaTab_5[i] = Accept; } 
  if ( len > 34 ) { nmea = st.substring(32,35); i = GetNmeaIndByStr(nmea); NmeaTab_5[i] = Accept; } 
  if ( len > 38 ) { nmea = st.substring(36,39); i = GetNmeaIndByStr(nmea); NmeaTab_5[i] = Accept; } 
  if ( len > 42 ) { nmea = st.substring(40,43); i = GetNmeaIndByStr(nmea); NmeaTab_5[i] = Accept; }  
}

void SetNmeaST1_5() {
 // this will change NmeaTab_5[]. This table has 0 (sentence is rejected) or 1 (sentence accepted).
 // we will go through the table and if you find a 0 we will do nothing. If we find a 1 we will 
 // write the index of the sentence instead of the 1. Indexes start with 1 and go up to 74. So no
 // sentence has 0 as an index!
  for ( int i = 0; i < NMEA183Size; i++) {
    if ( NmeaTab_5[i+1] == 1 ) { NmeaTab_5[i+1] = i+1; }
  }
}

void InitNmeaTab( int tab, int value) {
  for ( int i = 0; i < NMEA183Size; i++) {
    if ( tab == 1 ) { NmeaTab_1[i+1] = value; }
    if ( tab == 3 ) { NmeaTab_3[i+1] = value; }
    if ( tab == 4 ) { NmeaTab_4[i+1] = value; }
    if ( tab == 5 ) { NmeaTab_5[i+1] = value; }
  }
}

int GetNmeaIndByStr( String st ) {
    int a = st.charAt(0) & B00011111;
    int b = st.charAt(1) & B00011111;  
    int c = st.charAt(2) & B00011111; 
    a = a << 10;
    b = b << 5;
    a = a + b + c; 
    a = NmeaTable[a]; 
    return a; 
}

// this for P1 input
boolean CheckNmeaFlt_1() {
    if ( No_Flt_1 ) { return true; } 
    int a = nmea_1[3] & B00011111;
    int b = nmea_1[4] & B00011111; 
    int c = nmea_1[5] & B00011111;
    a = a << 10;
    b = b << 5;
    a = a + b + c; 
    a = NmeaTable[a];
    if ( NmeaTab_1[a] ) { return true; }
    return false; 
}

// this is for P3 output
boolean CheckNmeaFlt_3( int i ) {
  if ( No_Flt_3 ) { return true; }  
  int a, b, c;
  // i = 1 never happens in this bridge
  if ( i == 1 ) {
    a = nmea_1[3] & B00011111;
    b = nmea_1[4] & B00011111; 
    c = nmea_1[5] & B00011111; 
  }
  if ( i == 4 ) {
    a = nmea_4[3] & B00011111;
    b = nmea_4[4] & B00011111; 
    c = nmea_4[5] & B00011111; 
  }
  a = a << 10;
  b = b << 5;
  a = a + b + c; 
  a = NmeaTable[a];
  if ( NmeaTab_3[a] ) { return true; }
  return false; 
}

// this is for SeaTalk input
boolean CheckNmeaFlt_4() {
    if ( No_Flt_4 ) { return true; }  
    int a = nmea_4[3] & B00011111;
    int b = nmea_4[4] & B00011111; 
    int c = nmea_4[5] & B00011111;
    a = a << 10;
    b = b << 5;
    a = a + b + c; 
    a = NmeaTable[a];
    if ( NmeaTab_4[a] ) { return true; }
    return false; 
}

// this is for ST output; now returns a byte!
byte CheckNmeaFlt_5( int i ) {
  int a, b, c;
  // only i = 1 can happen in this bridge
  if ( i == 1 ) {
    a = nmea_1[3] & B00011111;
    b = nmea_1[4] & B00011111; 
    c = nmea_1[5] & B00011111; 
  }
  a = a << 10;
  b = b << 5;
  a = a + b + c; 
  a = NmeaTable[a];
  return NmeaTab_5[a]; 
}

boolean CheckFltStr(String st) {
  int len = st.length(); 
  String nmea;
  int i;
  if ( len == 1 ) {
    if ( st == "0" ) { return true; }
    if ( st == "1" ) { return true; }
    if ( st == "2" ) { return true; }
    if ( st == "3" ) { return true; }
    if ( st == "4" ) { return true; }
    if ( st == "5" ) { return true; }
    if ( st == "6" ) { return true; }
    if ( st == "7" ) { return true; }
    if ( st == "8" ) { return true; }
    if ( st == "9" ) { return true; }
  }
//  if ( len == 0 )  { return false; }  // never happens "" >>> "0"  
  nmea = st.substring(2,3); 
  if ( ( nmea != "A" ) && ( nmea != "R" ) ) { return false; } 
  if ( len >  6 ) { nmea = st.substring( 4,7 ); i = GetNmeaIndByStr(nmea); if (i == 0) { return false; } }
  if ( len > 10 ) { nmea = st.substring( 8,11); i = GetNmeaIndByStr(nmea); if (i == 0) { return false; } }
  if ( len > 14 ) { nmea = st.substring(12,15); i = GetNmeaIndByStr(nmea); if (i == 0) { return false; } }
  if ( len > 18 ) { nmea = st.substring(16,19); i = GetNmeaIndByStr(nmea); if (i == 0) { return false; } }
  if ( len > 22 ) { nmea = st.substring(20,23); i = GetNmeaIndByStr(nmea); if (i == 0) { return false; } }
  if ( len > 26 ) { nmea = st.substring(24,27); i = GetNmeaIndByStr(nmea); if (i == 0) { return false; } } 
  if ( len > 30 ) { nmea = st.substring(28,31); i = GetNmeaIndByStr(nmea); if (i == 0) { return false; } }
  if ( len > 34 ) { nmea = st.substring(32,35); i = GetNmeaIndByStr(nmea); if (i == 0) { return false; } }
  if ( len > 38 ) { nmea = st.substring(36,39); i = GetNmeaIndByStr(nmea); if (i == 0) { return false; } } 
  if ( len > 42 ) { nmea = st.substring(40,43); i = GetNmeaIndByStr(nmea); if (i == 0) { return false; } }  
  if ( len ==  7 ) { return true; } 
  if ( len == 11 ) { return true; }   
  if ( len == 15 ) { return true; } 
  if ( len == 19 ) { return true; }   
  if ( len == 23 ) { return true; } 
  if ( len == 27 ) { return true; }  
  if ( len == 31 ) { return true; } 
  if ( len == 35 ) { return true; }   
  if ( len == 39 ) { return true; } 
  if ( len == 43 ) { return true; }   
  return false;    
}
