/* =============================================
 * Bridge to convert SeaTalk1 to & from Nmea0183
 * =============================================
 *  
 *  Runs on an ESP32 Dev Module compiled with Arduino 1.8.10 and based on: 
 *  
 *  - SeaTalk Technical Reference by Thomas Knauf http://www.thomasknauf.de/seatalk.htm
 *  - NMEA Revealed by Eric S. Raymond http://www.catb.org/gpsd/NMEA.html
 *  
 *  This software a ESP32 module with 2 interfaces: a SeaTalk1 inteface and a Nmea0183 interface
 *  
 *  The SeaTalk1 interface uses 2 pins, TX_ST_PIN and RX_ST_PIN, to transmit and to receive                                              
 *  data. The connection to the SeaTalk1 bus can easily be built with a NPN transistor and
 *  4 resistors or a more elaborated one using a dual channel optocoupler.
 *  
 *  For circuit diagrams, reporting of errors, discussions, ... please visit:
 *  
 *  https://www.vela-navega.com/forum/viewforum.php?f=19
 *
 *  The Nmea0183 interface uses the hardware UART0 controller. To receive Nmea
 *  data we use Serial.read() and to to transmit we use Serial.print(). Therefore if the ESP
 *  module has a USB to Serial controller we can just use this USB port to send and to receive
 *  Nmea0183 sentences. See the variable Is_USB below. You can use this software freely and if
 *  you discover errors or add features, I would be grateful if you could point them to me.
 *  
 * Luis Sa - luis @ vela-navega . com on March, 28th, 2021
 * 
 */


// There are 3 variables that you can set for specially purposes and we describe them right now

boolean Is_USB = false;                // Setting Is_USB to true, the NMEA port will use GPIO pins 3 (for RX)
                                      // and 1 (for TX). Baud rate will be 115200 (can change that below)
                                      // and the USB port on the ESP32 module will be the NMEA port.
                                      // setting Is_USB to false, the NMEA port will use GPIO pins 26 (for RX)
                                      // and 1 (for TX). Baud rate will be 4800.

boolean DebugMode = false;            // Setting DebugMode to true the bridge outputs in the ST > NMEA
                                      // direction, not only the Nmea sentence that was converted from
                                      // the SeaTalk datagram but also the datagram(s) that originated
                                      // the sentence , For example you can get this:
                                      // ST=27 01 17 01
                                      // $IIMTW,17.9,C*1C
                                      
boolean LoopBack = true;              // Normally when you transmit data on the SeaTalk bus you stop
                                      // receiving as you do not need to receive what you are transmitting.
                                      // Setting this variable to true the reception on the ST bus is
                                      // mantained and if you send a sentence to the bridge, the sentence 
                                      // will be converted to a datagram which is read and converted back to
                                      // Nmea. You can use that by disconnecting the bridge from the 
                                      // SeaTalk bus to test conversions from NMEA > ST and from ST > NMEA
 

#define TX_ST_PIN    17                // GPIO17 drives the optocoupler LED that in turn drives the 
                                       // opto transistor. When GPIO17 is LOW the transistor is OFF
                                       // and SeaTalk DATA line is HIGH. When GPIO17 goes HIGH the
                                       // LED lights up and the transistor saturates bringing DATA line LOW       

#define RX_ST_PIN    18                // GPIO18 goes low when the DATA line goes low. It will generate
                                       // an interrupt and a timer routine will read the arriving ST datagram
                                       // Before the multiplexer starts transmission into the bus, the
                                       // reading of datagrams should be halted.
#define RX_ST_SIZE  256                // size of input buffer for Software Serial Uart                                                   

                                       // to store a SeaTalk datagram we can use 
char ST[18];                           // or byte ST[18] or uint8_t ST[18] which is used in ST_NMEA.INO file
                                       // array of bytes to contain a SeaTalk datagram. Maximum size is 18
                                       // The first byte ST[0] defines the type of datagram or the command.
                                       // For example ST[0]=0x00 means depth below transistor; ST[0]=0x26 means 
                                       // speed through water. We are referring to the elements of a datagram as bytes
                                       // but in fact each element has a Start Bit (0V) followed by 8 Data Bits.
                                       // (these are the bytes referred to above with least significant bit transmitted first).
                                       // Bits can be 0 (Data line=0V or txPin=HIGH) or 1 (Data line=12V or txPin=LOW).
                                       // Then there is a Sign Bit (the 9th bit) and a Stop Bit (12V). The first byte ST[0] has
                                       // a Sign Bit of 1 (12V). All the other bytes of the datagram have a Sign Bit
                                       // of 0 (0V).
                                       // Very Important are the 4 least significant bits of the 2nd byte. Adding 3 to
                                       // the number contained in this nibble gives the total lengh of the datagram.
                                       // For example 26 04 XX XX YY YY DE is a datagram of lenght 7 (speed over water) 
                                       // and, for example the ficticious datagram "3C 02 59 4A" is invalid as it
                                       // should have 5 bytes and only 4 are shown 

                                       // volatile variables should be used when interrupts need to process them
                                       // they will be placed in RAM and not in FLASH, which is much faster
volatile uint32_t bit_ticks;           // duration of a bit in 4800 baud in ticks, where each tick is 0.1 us 
volatile uint32_t init_ticks;          // ticks from falling edge to the sampling start

volatile uint32_t last_ST_time=0;      // the reception of a byte in the SeaTalk SeaTalk bus starts
                                       // when an interrupt is received (DATA line goes low). 
                                       // When the interrupt is received the time since power on
                                       // (in microsecoonds) is written here by the RX_UART_ST routine
                                       // Since each byte takes about 2100 microseconds to be received 
                                       // and the iddle bus time is also 2100 microseconds, the following
#define ST_wait 4200                   // is the waiting time in microseconds, before transmission starts 

// When a datagram is ready to be output it can be the case that we need to wait until the datagram is 
// effectively sent out. We need to build a circular buffer (FIFO) to accommodate for this. The buffer
// will have a length of 8. That means if datagrams are becoming ready to be sent out and we are not
// able to get the bus to send datagrams out, only the 9th datagram is lost. We think this is enough.
// We define the buffer as 8 blocks of length 18. 

volatile byte ST_buffer[8][19];     // we use 19 instead of 18 to count for length on 1st byre

// This is not the most efficient way (as datagrams rarely have its maximum length) but allows a fast and simple
// way to deal with the buffer. We can think this buffer as 8 blocks of 19 bytes each. The first byte
// of a block will have the lenght of the datagram. In a previous implementation the length of the
// datagram was determined by a null after the last byte. This was not possible as nulls can also be bytes of valid
// datagrams. For example if a datagram of length 7 is to be stored in block 1 then ST_buffer[1][0]=7. There is no
// need to buffer the 9th bit. When outputting block j (j = 0 1 2 3 4 5 6 or 7) we know that ST_buffer[j][1] will
// have a sign bit of 1 and the remaining bytes will have a sign bit of zero.

// We use 2 pointers to implement the circular buffer

volatile int inBuf = 0;         // recycling at 7. So they vary as: 0 1 2 3 0 1 2 3 4 5 6 7 0 1 2 ....
volatile int outBuf = 0;

/* The 2 pointers start as 0. We can not send data out if no data has come in. So when
  a timer interrupt checks for sending datagrams out, we look to these 2 pointers. If they are identical
  it means that there is no data to send out. Assume that 2 datagrams come and are placed in
  blocks 0 and 1, before any data is sent out. The pointers will be inBuf=2 and OutBuf=0. At this stage assume
  that we have an opportunity to send data out. We will send block 0 and then block 1 and will set OutBuf=2. So
  if we are asked for (prompted to) send more, we will do nothing because inBuf==outBuf (=2). Now suppose that 7 datagrams 
  arrive almost instantaneously. They will be placed in blocks 2 3 4 5 6 7 and 0. At this stage assume that we can send
  data out. Because inBuf is different from outBuf, data is sent out starting with the block pointed out by
  outBuf (which is 2) followed by 3 4 5 6 7 and by 0. Data out stops when outBuf "catches" inBuff. Let us assume
  that 2 more datagram are received and sent out. Both pointers will have the value of 2. And now consider
  that 10 datagrams are received before we get an opportunity to send anything out. We know that we can store
  the first 8 datagrams and that we will loose the 9th and the 10th. We know that the 1st datagram should
  go to block 2, the 2nd to block 3, the 3rd to block 4, ..., the  8th to block 1 and will loose the 9th and 10th
  datagrams. There is a simple way to implement this. When a datagram arrives we can have a procedure
  to receive datagrams (FILLING THE BUFFER):

      nextBuf = inBuf + 1; 
      if (nextBuf == 8) { nextBuf = 0; }
      if ( nextBuf == outBuf ) { return; }
      Store_ST(inBuf);
      inBuf = nextBuf;
      
  The procedure to Send_ST datagrams (EMPTYING THE BUFFER) will be:

     while ( inBuf != outBuf ) {
       Send_ST(outBuf);
       outBuf = outBuf + 1; 
       if (outBuf == 8) { outBuf = 0; }
     }
   
  However we can not do as shown. We need to relay on interrupts. When we add (FILL) a datagram to the buffer
  we start timer0 to interrupt periodically the main program. The bit time is 1/4800 seconds or 208 micro seconds.
  When we start timer0 we halt reception on the SeaTalk (we stop timer1). When we send out a complete datagram,
  we check if there are more to be sent. If there is no datagram to be sent we stop timer0 and we restart timer1.
  Just in case, we have the following check at the beginning of the onTimer0:

    if ( inBuf == outBuf ) { return; }

  If the test condition above is false we need to send out the datagram stored in the the block pointed out by outBuf.
  And we can only do so if the bus is being held in the idle state for a certain amount of time. We will do it like that

    // wait until bus is idle for at least ST_wait uS
    while ( ( micros() - last_ST_time ) < ST_wait ) {  
    Serial_ST_end();                                           // first thing to do is stop receiving interrupts on pin 18

    // following explanations can be found in TX_UART_ST.ino
 */

volatile boolean Output_ST = false;     // = false >>> no datagram output is needed, timer_0 is stopped (no interrupts)
                                        // = true >>> datagram was placed in the buffer, timer_0 started (interrupts)
                                    
volatile uint8_t txByte;                // we copy here the byte (read from the buffer) that needs to be transmitted
volatile uint8_t txSize;                // and here we copy its length

void IRAM_ATTR Output_ST_begin();       // should be in RAM

// variables related to the buffer for the incoming nmea data
boolean Led1 = false;               // used to blink led1
char nmea_1[83];                    // NMEA buffer for Port #1
int line_1 = 0;                     // data available flag for port #1
int i_1 = 0;                        // lenght of sentence on port  #1
boolean Isnmea_1 = false;           // clean Nmea sentence is being received on port #1

// variables related to the buffer for the converted (from ST) nmea data
boolean Led4 = false;                // used to blink led4
char nmea_4[150];                    // NMEA buffer for ST input port       // added 61 for ST preamble in Debug mode
int line_4 = 0;                      // data available flag for ST input port
int i_4 = 0;                         // lenght of sentence on ST port
boolean Is_ST = false;               // indicates that a "clean" SeaTalk datagram is being received

// filter strings                    // see description in Set_Filtering() or in FILTER.ino
char FltStr_1[43];                   // for input serial port P1 (NMEA in)
char FltStr_3[43];                   // for output serial port P3 (NMEA out)
char FltStr_4[43];                   // for input SeaTalk port
char FltStr_5[43];                   // for output SeaTalk port

// other variables used by the filter functions
int Skip_1 = 0; int FltSkip_1 ; boolean No_Flt_1, CSum_1;
int Skip_3 = 0; int FltSkip_3 ; boolean No_Flt_3, CSum_3;
int Skip_4 = 0; int FltSkip_4 ; boolean No_Flt_4, CSum_4;
int Skip_5 = 0; int FltSkip_5 ; boolean No_Flt_5, CSum_5;

void setup() {
  if ( Is_USB ) {
    Serial.begin(115200);
  }
  else {
    pinMode(26, INPUT_PULLUP);
    Serial.begin(4800, SERIAL_8N1, 26, 1);                 
  }
  // this is for filtering 
  Set_Filtering();
  // load ticks values
  bit_ticks = 10000000 / 4800;             // units = 0.1 micro seconds
  init_ticks = 1.4 * bit_ticks;
  // initial setup of both timers                   
  Serial_ST_setup();                       
  Output_ST_setup();
  // Start reception. Transmission will only be ON when needed!
  Serial_ST_begin();                        // for port#4 (SeaTalk input)
}

void loop() {
  yield();
  // this reads real SeaTalk data which is converted to NMEA (in nmea_4 buffer) and then send out by Output_4()
  Test_Serial_4();
  if (line_4) { Output_4(); }   

  // this reads NMEA data (stored in nmea_1 buffer) which is converted to ST datagrams and send out to the ST bus
  Test_Serial_1(); 
  if (line_1) { Output_1(); }
}

void Set_Filtering() {
  
/*  The input serial port, the output serial port, the ST input port and the ST output
    port can be set to pass or to block certain NMEA0183 sentences. This is governed by
    4 filtering strings of up to 43 characters that you can set for each of the referred
    to ports. The first character is a number, between 0 and 9. When the string is just
    0 no filtering takes place. When string starts with 9 no input or output happens.
    When the string starts with 1 the checksum of the nmea sentence is checked and
    the sentence is ignored if a bad checksum is found. If the string starts with
    a number from 2 up to 8, in addition to checksum validation only the 2nd up to
    the 8th sequence has a chance to pass the filter independently of the rest of the
    filtering string. For example, if that number is 4, it means that on 4 successive
    sequences, only the 4th is taken and 3 are dropped. The second character should be
    a separator. The characters "-", ".", "~" and others could be used as separators
    but we will use "-". The third character can only be "A" or "R". "A" means that only
    the sentences that are referred to in the rest of the filtering string are ACCEPTED by
    (pass through) the ports. "R" means that only the sentences that are referred to in
    the rest of the filtering string are REJECTED (blocked) by the ports. Each NMEA0183
    sequence is specified by 3 characters inserted between separators. For example, the string
    "2-A-RMC-GGA-GLL-GRS-GSA" means that only the 2nd in 2 successive sentences is considered
    and the considered sequence only passes if it is 1 of the 5 that are specified.
    It is interesting to note that the shown sequence has a length of 23 characters and 
    specifies 5 sentences. For each additional sentence you need 4 more characters
    (the separator and the 3 characters). Therefore, in order to specify 10 sentences we need
    43 characters which is maximum length of the filtering string. 
 */  

  // start by building tables
  BuildNmeaTables();  // needed before calling CheckFltStr(String st)

  // you can change the following filtering strings. If they have not the correct sintax,
  // they become switch back to the default "0" (meaning: no filtering)
  String str_1 = "0";               // for input serial port P1
  String str_3 = "0";               // for output serial port P3
  String str_4 = "0";               // for input SeaTalk port
  String str_5 = "0";               // for output SeaTalk port

  // now check if the strings are OK
  if ( CheckFltStr(str_1) == false ) { Serial.println("Incorrect Filter 1 String"); str_1 = "0"; }
  str_1.toCharArray(FltStr_1, str_1.length() + 1);

  if ( CheckFltStr(str_3) == false ) { Serial.println("Incorrect Filter 3 String"); str_3 = "0"; }
  str_3.toCharArray(FltStr_3, str_3.length() + 1);
  
  if ( CheckFltStr(str_4) == false ) { Serial.println("Incorrect Filter 4 String"); str_4 = "0"; }
  str_4.toCharArray(FltStr_4, str_4.length() + 1);
  
  if ( CheckFltStr(str_5) == false ) { Serial.println("Incorrect Filter 5 String"); str_5 = "0"; }
  str_5.toCharArray(FltStr_5, str_5.length() + 1);

  // now using filter strings set fiter tables
  SetNmeaTab_1();
  SetNmeaTab_3();
  SetNmeaTab_4();
  SetNmeaTab_5();
  SetNmeaST1_5();     // this one after SetNmeaTab_5()
  BuildHexTable();    // needed for checksum verification
  

  No_Flt_1 = false; No_Flt_3 = false; No_Flt_4 = false; No_Flt_5 = false;
  if ( str_1 == "0" ) { No_Flt_1 = true; }
  if ( str_3 == "0" ) { No_Flt_3 = true; }
  if ( str_4 == "0" ) { No_Flt_4 = true; }  
  if ( str_5 == "0" ) { No_Flt_5 = true; }

  CSum_1 = true; CSum_3 = true; CSum_4 = true; CSum_5 = true;
  if ( FltStr_1[0] == 48 ) { CSum_1 = false; }
  if ( FltStr_3[0] == 48 ) { CSum_3 = false; } 
  if ( FltStr_4[0] == 48 ) { CSum_4 = false; }
  if ( FltStr_5[0] == 48 ) { CSum_5 = false; }
  
}
